from .config import (
    DAY,
    ETERNITY,
    INSTANT_PATTERN,
    MONTH,
    YEAR,
    date_by_instant_cache,
    str_by_instant_cache,
    year_or_month_or_day_re,
)
from .helpers import (
    N_,
    instant,
    instant_date,
    key_period_size,
    period,
    unit_weight,
    unit_weights,
)
from .instant_ import Instant
from .period_ import Period



"""
This file defines variables for the modelled legislation.

A variable is a property of an Entity such as a Person, a Household…

See https://openfisca.org/doc/key-concepts/variables.html
"""

# Import from numpy the operations you need to apply on OpenFisca's population vectors
# Import from openfisca-core the Python objects used to code the legislation in OpenFisca
from numpy import maximum as max_

# Import the Entities specifically defined for this tax and benefit system
from policyengine_core.country_template.entities import Household, Person
from policyengine_core.periods import MONTH, YEAR
from policyengine_core.variables import Variable


class income_tax(Variable):
    value_type = float
    entity = Person
    definition_period = MONTH
    label = "Income tax"
    reference = "https://law.gov.example/income_tax"  # Always use the most official source

    def formula(person, period, parameters):
        """
        Income tax.

        The formula to compute the income tax for a given person at a given period
        """
        return (
            person("salary", period) * parameters(period).taxes.income_tax_rate
        )


class social_security_contribution(Variable):
    value_type = float
    entity = Person
    definition_period = MONTH
    label = (
        "Progressive contribution paid on salaries to finance social security"
    )
    reference = "https://law.gov.example/social_security_contribution"  # Always use the most official source

    def formula(person, period, parameters):
        """
        Social security contribution.

        The social_security_contribution is computed according to a marginal scale.
        """
        salary = person("salary", period)
        scale = parameters(period).taxes.social_security_contribution

        return scale.calc(salary)


class housing_tax(Variable):
    value_type = float
    entity = Household
    definition_period = YEAR  # This housing tax is defined for a year.
    label = "Tax paid by each household proportionally to the size of its accommodation"
    reference = "https://law.gov.example/housing_tax"  # Always use the most official source

    def formula(household, period, parameters):
        """
        Housing tax.

        The housing tax is defined for a year, but depends on the `accommodation_size` and `housing_occupancy_status` on the first month of the year.
        Here period is a year. We can get the first month of a year with the following shortcut.
        To build different periods, see https://openfisca.org/doc/coding-the-legislation/35_periods.html#calculate-dependencies-for-a-specific-period
        """
        january = period.first_month
        accommodation_size = household("accommodation_size", january)

        tax_params = parameters(period).taxes.housing_tax
        tax_amount = max_(
            accommodation_size * tax_params.rate, tax_params.minimal_amount
        )

        # `housing_occupancy_status` is an Enum variable
        occupancy_status = household("housing_occupancy_status", january)
        HousingOccupancyStatus = (
            occupancy_status.possible_values
        )  # Get the enum associated with the variable
        # To access an enum element, we use the `.` notation.
        tenant = occupancy_status == HousingOccupancyStatus.tenant
        owner = occupancy_status == HousingOccupancyStatus.owner

        # The tax is applied only if the household owns or rents its main residency
        return (owner + tenant) * tax_amount



class ParameterNotFoundError(AttributeError):
    """
    Exception raised when a parameter is not found in the parameters.
    """

    def __init__(self, name, instant_str, variable_name=None):
        """
        :param name: Name of the parameter
        :param instant_str: Instant where the parameter does not exist, in the format `YYYY-MM-DD`.
        :param variable_name: If the parameter was queried during the computation of a variable, name of that variable.
        """
        self.name = name
        self.instant_str = instant_str
        self.variable_name = variable_name
        message = "The parameter '{}'".format(name)
        if variable_name is not None:
            message += " requested by variable '{}'".format(variable_name)
        message += (" was not found in the {} tax and benefit system.").format(
            instant_str
        )
        super(ParameterNotFoundError, self).__init__(message)



from .get_parameter import get_parameter
from .homogenize_parameters import homogenize_parameter_structures
from .interpolate_parameters import interpolate_parameters
from .propagate_parameter_metadata import propagate_parameter_metadata
from .uprate_parameters import uprate_parameters



class PeriodMismatchError(ValueError):
    """
    Exception raised when one tries to set a variable value for a period that doesn't match its definition period
    """

    def __init__(self, variable_name, period, definition_period, message):
        self.variable_name = variable_name
        self.period = period
        self.definition_period = definition_period
        self.message = message
        ValueError.__init__(self, self.message)



import os
import textwrap
from typing import Any

from policyengine_core.entities.role import Role


class Entity:
    """
    Represents an entity (e.g. a person, a household, etc.) on which calculations can be run.
    """

    def __init__(self, key: str, plural: str, label: str, doc: str):
        self.key = key
        self.label = label
        self.plural = plural
        self.doc = textwrap.dedent(doc)
        self.is_person = True
        self._tax_benefit_system = None

    def set_tax_benefit_system(self, tax_benefit_system) -> None:
        self._tax_benefit_system = tax_benefit_system

    def check_role_validity(self, role: Any) -> None:
        if role is not None and not type(role) == Role:
            raise ValueError("{} is not a valid role".format(role))

    def get_variable(self, variable_name: str, check_existence: bool = False):
        return self._tax_benefit_system.get_variable(
            variable_name, check_existence
        )

    def check_variable_defined_for_entity(self, variable_name: str) -> None:
        variable_entity = self.get_variable(
            variable_name, check_existence=True
        ).entity
        # Should be this:
        # if variable_entity is not self:
        if variable_entity.key != self.key:
            message = os.linesep.join(
                [
                    "You tried to compute the variable '{0}' for the entity '{1}';".format(
                        variable_name, self.plural
                    ),
                    "however the variable '{0}' is defined for '{1}'.".format(
                        variable_name, variable_entity.plural
                    ),
                    "Learn more about entities in our documentation:",
                    "<https://openfisca.org/doc/coding-the-legislation/50_entities.html>.",
                ]
            )
            raise ValueError(message)



"""
This file defines variables for the modelled legislation.

A variable is a property of an Entity such as a Person, a Household…

See https://openfisca.org/doc/key-concepts/variables.html
"""

# Import the Entities specifically defined for this tax and benefit system
from policyengine_core.country_template.entities import Person

# Import from openfisca-core the Python objects used to code the legislation in OpenFisca
from policyengine_core.holders import set_input_divide_by_period
from policyengine_core.periods import MONTH
from policyengine_core.variables import Variable


# This variable is a pure input: it doesn't have a formula
class salary(Variable):
    value_type = float
    entity = Person
    definition_period = MONTH
    set_input = set_input_divide_by_period  # Optional attribute. Allows user to declare a salary for a year. OpenFisca will spread the yearly amount over the months contained in the year.
    label = "Salary"
    reference = (
        "https://law.gov.example/salary"  # Always use the most official source
    )


class disposable_income(Variable):
    value_type = float
    entity = Person
    definition_period = MONTH
    label = "Actual amount available to the person at the end of the month"
    reference = "https://stats.gov.example/disposable_income"  # Some variables represent quantities used in economic models, and not defined by law. Always give the source of your definitions.

    def formula(person, period, _parameters):
        """Disposable income."""
        return (
            +person("salary", period)
            + person("basic_income", period)
            - person("income_tax", period)
            - person("social_security_contribution", period)
        )



from policyengine_core.errors import SituationParsingError


def calculate_output_add(simulation, variable_name, period):
    return simulation.calculate_add(variable_name, period)


def calculate_output_divide(simulation, variable_name, period):
    return simulation.calculate_divide(variable_name, period)


def check_type(input, input_type, path=None):
    json_type_map = {
        dict: "Object",
        list: "Array",
        str: "String",
    }

    if path is None:
        path = []

    if not isinstance(input, input_type):
        raise SituationParsingError(
            path,
            "Invalid type: must be of type '{}'.".format(
                json_type_map[input_type]
            ),
        )


def transform_to_strict_syntax(data):
    if isinstance(data, (str, int)):
        data = [data]
    if isinstance(data, list):
        return [str(item) if isinstance(item, int) else item for item in data]
    return data


def _get_person_count(input_dict):
    try:
        first_value = next(iter(input_dict.values()))
        if isinstance(first_value, dict):
            first_value = next(iter(first_value.values()))
        if isinstance(first_value, str):
            return 1

        return len(first_value)
    except Exception:
        return 1



import textwrap


class Role:
    """
    The type of the relation between an entity instance and a group entity instance.
    """

    def __init__(self, description, entity):
        self.entity = entity
        self.key = description["key"]
        self.label = description.get("label")
        self.plural = description.get("plural")
        self.doc = textwrap.dedent(description.get("doc", ""))
        self.max = description.get("max")
        self.subroles = None

    def __repr__(self) -> str:
        return "Role({})".format(self.key)



from policyengine_core.projectors import Projector


class UniqueRoleToEntityProjector(Projector):
    """For instance famille.declarant_principal."""

    def __init__(self, entity, role, parent=None):
        self.target_entity = entity
        self.reference_entity = entity.members
        self.parent = parent
        self.role = role

    def transform(self, result):
        return self.target_entity.value_from_person(result, self.role)



import plotly.graph_objects as go
from IPython.display import HTML


WHITE = "#FFF"
LIGHTER_BLUE = "#ABCEEB"  # Blue 100.
LIGHT_BLUE = "#49A6E2"  # Blue 500.
BLUE = "#1976D2"  # Blue 700.
DARK_BLUE = "#0F4AA1"  # Blue 900.
GRAY = "#BDBDBD"
DARK_GRAY = "#616161"
LIGHT_GRAY = "#F5F5F5"
LIGHT_GREEN = "#C5E1A5"
DARK_GREEN = "#558B2F"


def format_fig(fig: go.Figure) -> dict:
    """Formats figure with styling and returns as JSON.
    :param fig: Plotly figure.
    :type fig: go.Figure
    :return: Formatted plotly figure as a JSON dict.
    :rtype: dict
    """
    fig.update_xaxes(
        title_font=dict(size=16, color="black"), tickfont={"size": 14}
    )
    fig.update_yaxes(
        title_font=dict(size=16, color="black"), tickfont={"size": 14}
    )
    return fig.update_layout(
        hoverlabel_align="right",
        font_family="Arial, sans-serif",
        font_color="Black",
        title_font_size=20,
        plot_bgcolor="white",
        paper_bgcolor="white",
        hoverlabel=dict(font_family="Arial, sans-serif"),
    )


def display_fig(fig: go.Figure) -> HTML:
    return HTML(fig.to_html(full_html=False, include_plotlyjs="cdn"))



class Unit:
    """The real-world meaning of a particular value."""

    GBP = "currency-GBP"
    """British pounds sterling."""

    USD = "currency-USD"
    """United States dollars."""

    EUR = "currency-EUR"
    """Euros."""

    PERCENT = "/1"
    """A percentage."""



from __future__ import annotations

import logging
import typing

from policyengine_core import taxscales

log = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from policyengine_core.parameters import ParameterNodeAtInstant

    TaxScales = typing.Optional[taxscales.MarginalRateTaxScale]


def combine_tax_scales(
    node: ParameterNodeAtInstant,
    combined_tax_scales: TaxScales = None,
) -> TaxScales:
    """
    Combine all the MarginalRateTaxScales in the node into a single
    MarginalRateTaxScale.
    """

    name = next(iter(node or []), None)

    if name is None:
        return combined_tax_scales

    if combined_tax_scales is None:
        combined_tax_scales = taxscales.MarginalRateTaxScale(name=name)
        combined_tax_scales.add_bracket(0, 0)

    for child_name in node:
        child = node[child_name]

        if isinstance(child, taxscales.MarginalRateTaxScale):
            combined_tax_scales.add_tax_scale(child)

        else:
            log.info(
                f"Skipping {child_name} with value {child} "
                "because it is not a marginal rate tax scale",
            )

    return combined_tax_scales



from __future__ import annotations

import time
import typing
from typing import Dict, Iterator, List, Optional, Union

from .. import tracers

if typing.TYPE_CHECKING:
    from numpy.typing import ArrayLike

    from policyengine_core.periods import Period

    Stack = List[Dict[str, Union[str, Period]]]


class FullTracer:

    _simple_tracer: tracers.SimpleTracer
    _trees: list
    _current_node: Optional[tracers.TraceNode]

    def __init__(self) -> None:
        self._simple_tracer = tracers.SimpleTracer()
        self._trees = []
        self._current_node = None

    def record_calculation_start(
        self,
        variable: str,
        period: Period,
        branch_name: str = "default",
    ) -> None:
        self._simple_tracer.record_calculation_start(
            variable, period, branch_name
        )
        self._enter_calculation(variable, period, branch_name)
        self._record_start_time()

    def _enter_calculation(
        self,
        variable: str,
        period: Period,
        branch_name: str = "default",
    ) -> None:
        new_node = tracers.TraceNode(
            name=variable,
            period=period,
            parent=self._current_node,
            branch_name=branch_name,
        )

        if self._current_node is None:
            self._trees.append(new_node)

        else:
            self._current_node.append_child(new_node)

        self._current_node = new_node

    def record_parameter_access(
        self,
        parameter: str,
        period: Period,
        branch_name: str,
        value: ArrayLike,
    ) -> None:

        if self._current_node is not None:
            self._current_node.parameters.append(
                tracers.TraceNode(
                    name=parameter,
                    period=period,
                    branch_name=branch_name,
                    value=value,
                ),
            )

    def _record_start_time(
        self,
        time_in_s: Optional[float] = None,
    ) -> None:
        if time_in_s is None:
            time_in_s = self._get_time_in_sec()

        if self._current_node is not None:
            self._current_node.start = time_in_s

    def record_calculation_result(self, value: ArrayLike) -> None:
        if self._current_node is not None:
            self._current_node.value = value

    def record_calculation_end(self) -> None:
        self._simple_tracer.record_calculation_end()
        self._record_end_time()
        self._exit_calculation()

    def _record_end_time(
        self,
        time_in_s: Optional[float] = None,
    ) -> None:
        if time_in_s is None:
            time_in_s = self._get_time_in_sec()

        if self._current_node is not None:
            self._current_node.end = time_in_s

    def _exit_calculation(self) -> None:
        if self._current_node is not None:
            self._current_node = self._current_node.parent

    @property
    def stack(self) -> Stack:
        return self._simple_tracer.stack

    @property
    def trees(self) -> List[tracers.TraceNode]:
        return self._trees

    @property
    def computation_log(self) -> tracers.ComputationLog:
        return tracers.ComputationLog(self)

    @property
    def performance_log(self) -> tracers.PerformanceLog:
        return tracers.PerformanceLog(self)

    @property
    def flat_trace(self) -> tracers.FlatTrace:
        return tracers.FlatTrace(self)

    def _get_time_in_sec(self) -> float:
        return time.time_ns() / (10**9)

    def print_computation_log(self, aggregate=False, max_depth=None):
        self.computation_log.print_log(aggregate, max_depth)

    def generate_performance_graph(self, dir_path: str) -> None:
        self.performance_log.generate_graph(dir_path)

    def generate_performance_tables(self, dir_path: str) -> None:
        self.performance_log.generate_performance_tables(dir_path)

    def _get_nb_requests(self, tree: tracers.TraceNode, variable: str) -> int:
        tree_call = tree.name == variable
        children_calls = sum(
            self._get_nb_requests(child, variable) for child in tree.children
        )

        return tree_call + children_calls

    def get_nb_requests(self, variable: str) -> int:
        return sum(
            self._get_nb_requests(tree, variable) for tree in self.trees
        )

    def get_flat_trace(self) -> dict:
        return self.flat_trace.get_trace()

    def get_serialized_flat_trace(self) -> dict:
        return self.flat_trace.get_serialized_trace()

    def browse_trace(self) -> Iterator[tracers.TraceNode]:
        def _browse_node(node):
            yield node

            for child in node.children:
                yield from _browse_node(child)

        for node in self._trees:
            yield from _browse_node(node)



import datetime
import inspect
import re
import textwrap
from typing import Callable, List, Type

import numpy
import sortedcontainers

from policyengine_core import periods, tools
from policyengine_core.entities import Entity
from policyengine_core.enums import Enum, EnumArray
from policyengine_core.periods import Period

from . import config, helpers


class QuantityType:
    STOCK = "stock"
    FLOW = "flow"


class Variable:
    """
    A `variable <https://openfisca.org/doc/key-concepts/variables.html>`_ of the legislation.
    """

    name: str
    """Name of the variable"""

    value_type: type
    """The value type of the variable. Possible value types in OpenFisca are ``int`` ``float`` ``bool`` ``str`` ``date`` and ``Enum``."""

    entity: Entity
    """`Entity <https://openfisca.org/doc/key-concepts/person,_entities,_role.html>`_ the variable is defined for. For instance : ``Person``, ``Household``."""

    definition_period: Period
    """`Period <https://openfisca.org/doc/coding-the-legislation/35_periods.html>`_ the variable is defined for. Possible value: ``MONTH``, ``YEAR``, ``ETERNITY``."""

    formulas: List[Callable]
    """Formulas used to calculate the variable"""

    label: str
    """Description of the variable"""

    reference: str
    """Legislative reference describing the variable."""

    default_value: object
    """`Default value <https://openfisca.org/doc/key-concepts/variables.html#default-values>`_ of the variable."""

    baseline_variable: str
    """If the variable has been introduced in a `reform <https://openfisca.org/doc/key-concepts/reforms.html>`_ to replace another variable, baseline_variable is the replaced variable."""

    dtype: numpy.dtype
    """Numpy `dtype <https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.dtype.html>`_ used under the hood for the variable."""

    end: datetime.date
    """`Date <https://openfisca.org/doc/coding-the-legislation/40_legislation_evolutions.html#variable-end>`_  when the variable disappears from the legislation."""

    is_neutralized: bool
    """True if the variable is neutralized. Neutralized variables never use their formula, and only return their default values when calculated."""

    json_type: str
    """JSON type corresponding to the variable."""

    max_length: int
    """If the value type of the variable is ``str``, max length of the string allowed. ``None`` if there is no limit."""

    possible_values: EnumArray
    """If the value type of the variable is ``Enum``, contains the values the variable can take."""

    set_input: Callable
    """Function used to automatically process variable inputs defined for periods not matching the definition_period of the variable. See more on the `documentation <https://openfisca.org/doc/coding-the-legislation/35_periods.html#set-input-automatically-process-variable-inputs-defined-for-periods-not-matching-the-definition-period>`_. Possible values are ``set_input_dispatch_by_period``, ``set_input_divide_by_period``, or nothing."""

    unit: str
    """Free text field describing the unit of the variable. Only used as metadata."""

    documentation: str
    """Free multilines text field describing the variable context and usage."""

    quantity_type: str
    """Categorical attribute describing whether the variable is a stock or a flow."""

    defined_for: str = None
    """The name of another variable, nonzero values of which are used to define the set of entities for which this variable is defined."""

    def __init__(self, baseline_variable=None):
        self.name = self.__class__.__name__
        attr = {
            name: value
            for name, value in self.__class__.__dict__.items()
            if not name.startswith("__")
        }
        self.baseline_variable = baseline_variable
        self.value_type = self.set(
            attr,
            "value_type",
            required=True,
            allowed_values=config.VALUE_TYPES.keys(),
        )
        self.dtype = config.VALUE_TYPES[self.value_type]["dtype"]
        self.json_type = config.VALUE_TYPES[self.value_type]["json_type"]
        if self.value_type == Enum:
            self.possible_values: Type[Enum] = self.set(
                attr,
                "possible_values",
                required=True,
                setter=self.set_possible_values,
            )
        if self.value_type == str:
            self.max_length = self.set(attr, "max_length", allowed_type=int)
            if self.max_length:
                self.dtype = "|S{}".format(self.max_length)
        if self.value_type == Enum:
            self.default_value = self.set(
                attr,
                "default_value",
                allowed_type=self.possible_values,
                required=True,
            )
        else:
            self.default_value = self.set(
                attr,
                "default_value",
                allowed_type=self.value_type,
                default=config.VALUE_TYPES[self.value_type].get("default"),
            )
        self.entity = self.set(
            attr, "entity", required=True, setter=self.set_entity
        )
        self.definition_period = self.set(
            attr,
            "definition_period",
            required=True,
            allowed_values=(
                periods.DAY,
                periods.MONTH,
                periods.YEAR,
                periods.ETERNITY,
            ),
        )
        self.quantity_type = self.set(
            attr,
            "quantity_type",
            required=False,
            allowed_values=(QuantityType.STOCK, QuantityType.FLOW),
            default=QuantityType.FLOW,
        )
        self.label = self.set(
            attr, "label", allowed_type=str, setter=self.set_label
        )
        self.end = self.set(attr, "end", allowed_type=str, setter=self.set_end)
        self.reference = self.set(attr, "reference", setter=self.set_reference)
        self.cerfa_field = self.set(
            attr, "cerfa_field", allowed_type=(str, dict)
        )
        self.unit = self.set(attr, "unit", allowed_type=str)
        self.documentation = self.set(
            attr,
            "documentation",
            allowed_type=str,
            setter=self.set_documentation,
        )
        self.set_input = self.set_set_input(attr.pop("set_input", None))
        self.calculate_output = self.set_calculate_output(
            attr.pop("calculate_output", None)
        )
        self.is_period_size_independent = self.set(
            attr,
            "is_period_size_independent",
            allowed_type=bool,
            default=config.VALUE_TYPES[self.value_type][
                "is_period_size_independent"
            ],
        )

        self.defined_for = self.set_defined_for(attr.pop("defined_for", None))

        formulas_attr, unexpected_attrs = helpers._partition(
            attr,
            lambda name, value: name.startswith(config.FORMULA_NAME_PREFIX),
        )
        self.formulas = self.set_formulas(formulas_attr)

        if unexpected_attrs:
            raise ValueError(
                'Unexpected attributes in definition of variable "{}": {!r}'.format(
                    self.name, ", ".join(sorted(unexpected_attrs.keys()))
                )
            )

        self.is_neutralized = False

    # ----- Setters used to build the variable ----- #

    def set(
        self,
        attributes,
        attribute_name,
        required=False,
        allowed_values=None,
        allowed_type=None,
        setter=None,
        default=None,
    ):
        value = attributes.pop(attribute_name, None)
        if value is None and self.baseline_variable:
            return getattr(self.baseline_variable, attribute_name)
        if required and value is None:
            raise ValueError(
                "Missing attribute '{}' in definition of variable '{}'.".format(
                    attribute_name, self.name
                )
            )
        if (
            required
            and allowed_values is not None
            and value not in allowed_values
        ):
            raise ValueError(
                "Invalid value '{}' for attribute '{}' in variable '{}'. Allowed values are '{}'.".format(
                    value, attribute_name, self.name, allowed_values
                )
            )
        if (
            allowed_type is not None
            and value is not None
            and not isinstance(value, allowed_type)
        ):
            if allowed_type == float and isinstance(value, int):
                value = float(value)
            else:
                raise ValueError(
                    "Invalid value '{}' for attribute '{}' in variable '{}'. Must be of type '{}'.".format(
                        value, attribute_name, self.name, allowed_type
                    )
                )
        if setter is not None:
            value = setter(value)
        if value is None and default is not None:
            return default
        return value

    def set_entity(self, entity):
        if not isinstance(entity, Entity):
            raise ValueError(
                f"Invalid value '{entity}' for attribute 'entity' in variable '{self.name}'. Must be an instance of Entity."
            )
        return entity

    def set_possible_values(self, possible_values):
        if not issubclass(possible_values, Enum):
            raise ValueError(
                "Invalid value '{}' for attribute 'possible_values' in variable '{}'. Must be a subclass of {}.".format(
                    possible_values, self.name, Enum
                )
            )
        return possible_values

    def set_label(self, label):
        if label:
            return label

    def set_end(self, end):
        if end:
            try:
                return datetime.datetime.strptime(end, "%Y-%m-%d").date()
            except ValueError:
                raise ValueError(
                    "Incorrect 'end' attribute format in '{}'. 'YYYY-MM-DD' expected where YYYY, MM and DD are year, month and day. Found: {}".format(
                        self.name, end
                    )
                )

    def set_reference(self, reference):
        if reference:
            if isinstance(reference, str):
                reference = [reference]
            elif isinstance(reference, list):
                pass
            elif isinstance(reference, tuple):
                reference = list(reference)
            elif isinstance(reference, dict):
                reference = [reference]

        return reference

    def set_documentation(self, documentation):
        if documentation:
            return textwrap.dedent(documentation)

    def set_set_input(self, set_input):
        if not set_input and self.baseline_variable:
            return self.baseline_variable.set_input
        return set_input

    def set_calculate_output(self, calculate_output):
        if not calculate_output and self.baseline_variable:
            return self.baseline_variable.calculate_output
        return calculate_output

    def set_formulas(self, formulas_attr):
        formulas = sortedcontainers.sorteddict.SortedDict()
        for formula_name, formula in formulas_attr.items():
            starting_date = self.parse_formula_name(formula_name)

            if self.end is not None and starting_date > self.end:
                raise ValueError(
                    'You declared that "{}" ends on "{}", but you wrote a formula to calculate it from "{}" ({}). The "end" attribute of a variable must be posterior to the start dates of all its formulas.'.format(
                        self.name, self.end, starting_date, formula_name
                    )
                )

            formulas[str(starting_date)] = formula

        # If the variable is reforming a baseline variable, keep the formulas from the latter when they are not overridden by new formulas.
        if self.baseline_variable is not None:
            first_reform_formula_date = (
                formulas.peekitem(0)[0] if formulas else None
            )
            formulas.update(
                {
                    baseline_start_date: baseline_formula
                    for baseline_start_date, baseline_formula in self.baseline_variable.formulas.items()
                    if first_reform_formula_date is None
                    or baseline_start_date < first_reform_formula_date
                }
            )

        return formulas

    def set_defined_for(self, defined_for):
        if isinstance(defined_for, Enum):
            defined_for = defined_for.value
        return defined_for

    def parse_formula_name(self, attribute_name):
        """
        Returns the starting date of a formula based on its name.

        Valid dated name formats are : 'formula', 'formula_YYYY', 'formula_YYYY_MM' and 'formula_YYYY_MM_DD' where YYYY, MM and DD are a year, month and day.

        By convention, the starting date of:
            - `formula` is `0001-01-01` (minimal date in Python)
            - `formula_YYYY` is `YYYY-01-01`
            - `formula_YYYY_MM` is `YYYY-MM-01`
        """

        def raise_error():
            raise ValueError(
                'Unrecognized formula name in variable "{}". Expecting "formula_YYYY" or "formula_YYYY_MM" or "formula_YYYY_MM_DD where YYYY, MM and DD are year, month and day. Found: "{}".'.format(
                    self.name, attribute_name
                )
            )

        if attribute_name == config.FORMULA_NAME_PREFIX:
            return datetime.date.min

        FORMULA_REGEX = r"formula_(\d{4})(?:_(\d{2}))?(?:_(\d{2}))?$"  # YYYY or YYYY_MM or YYYY_MM_DD

        match = re.match(FORMULA_REGEX, attribute_name)
        if not match:
            raise_error()
        date_str = "-".join(
            [match.group(1), match.group(2) or "01", match.group(3) or "01"]
        )

        try:
            return datetime.datetime.strptime(date_str, "%Y-%m-%d").date()
        except ValueError:  # formula_2005_99_99 for instance
            raise_error()

    # ----- Methods ----- #

    def is_input_variable(self):
        """
        Returns True if the variable is an input variable.
        """
        return len(self.formulas) == 0

    @classmethod
    def get_introspection_data(cls, tax_benefit_system):
        """
        Get instrospection data about the code of the variable.

        :returns: (comments, source file path, source code, start line number)
        :rtype: tuple

        """
        comments = inspect.getcomments(cls)

        # Handle dynamically generated variable classes or Jupyter Notebooks, which have no source.
        try:
            absolute_file_path = inspect.getsourcefile(cls)
        except TypeError:
            source_file_path = None
        else:
            source_file_path = absolute_file_path.replace(
                tax_benefit_system.get_package_metadata()["location"], ""
            )
        try:
            source_lines, start_line_number = inspect.getsourcelines(cls)
            source_code = textwrap.dedent("".join(source_lines))
        except (IOError, TypeError):
            source_code, start_line_number = None, None

        return comments, source_file_path, source_code, start_line_number

    def get_formula(self, period=None):
        """
        Returns the formula used to compute the variable at the given period.

        If no period is given and the variable has several formula, return the oldest formula.

        :returns: Formula used to compute the variable

        """

        if not self.formulas:
            return None

        if period is None:
            return self.formulas.peekitem(index=0)[
                1
            ]  # peekitem gets the 1st key-value tuple (the oldest start_date and formula). Return the formula.

        if isinstance(period, Period):
            instant = period.start
        else:
            try:
                instant = periods.period(period).start
            except ValueError:
                instant = periods.instant(period)

        if self.end and instant.date > self.end:
            return None

        instant = str(instant)
        for start_date in reversed(self.formulas):
            if start_date <= instant:
                return self.formulas[start_date]

        return None

    def clone(self):
        clone = self.__class__()
        return clone

    def check_set_value(self, value):
        if self.value_type == Enum and isinstance(value, str):
            try:
                value = self.possible_values[value].index
            except KeyError:
                possible_values = [item.name for item in self.possible_values]
                raise ValueError(
                    "'{}' is not a known value for '{}'. Possible values are ['{}'].".format(
                        value, self.name, "', '".join(possible_values)
                    )
                )
        if self.value_type in (float, int) and isinstance(value, str):
            try:
                value = tools.eval_expression(value)
            except SyntaxError:
                raise ValueError(
                    "I couldn't understand '{}' as a value for '{}'".format(
                        value, self.name
                    )
                )

        try:
            value = numpy.array([value], dtype=self.dtype)[0]
        except (TypeError, ValueError):
            if self.value_type == datetime.date:
                error_message = "Can't deal with date: '{}'.".format(value)
            else:
                error_message = "Can't deal with value: expected type {}, received '{}'.".format(
                    self.json_type, value
                )
            raise ValueError(error_message)
        except (OverflowError):
            error_message = "Can't deal with value: '{}', it's too large for type '{}'.".format(
                value, self.json_type
            )
            raise ValueError(error_message)

        return value

    def default_array(self, array_size):
        array = numpy.empty(array_size, dtype=self.dtype)
        if self.value_type == Enum:
            array.fill(self.default_value.index)
            return EnumArray(array, self.possible_values)
        array.fill(self.default_value)
        return array



import os

import dpath.util


class SituationParsingError(Exception):
    """
    Exception raised when the situation provided as an input for a simulation cannot be parsed
    """

    def __init__(self, path, message, code=None):
        self.error = {}
        dpath_path = "/".join([str(item) for item in path])
        message = str(message).strip(os.linesep).replace(os.linesep, " ")
        dpath.util.new(self.error, dpath_path, message)
        self.code = code
        Exception.__init__(self, str(self.error))

    def __str__(self):
        return str(self.error)



import importlib
import logging
import pkgutil
import traceback
from os import linesep

log = logging.getLogger(__name__)


def add_tax_benefit_system_arguments(parser, country_only: bool = False):
    parser.add_argument(
        "-c",
        "--country-package",
        action="store",
        help='country package to use. If not provided, an automatic detection will be attempted by scanning the python packages installed in your environment which name contains the word "policyengine".',
    )
    if country_only:
        return parser
    parser.add_argument(
        "-e",
        "--extensions",
        action="store",
        help="extensions to load",
        nargs="*",
    )
    parser.add_argument(
        "-r",
        "--reforms",
        action="store",
        help="reforms to apply to the country package",
        nargs="*",
    )

    return parser


def build_tax_benefit_system(country_package_name, extensions, reforms):
    if country_package_name is None:
        country_package_name = detect_country_package()
    try:
        country_package = importlib.import_module(country_package_name)
    except ImportError:
        message = linesep.join(
            [
                traceback.format_exc(),
                "Could not import module `{}`.".format(country_package_name),
                "Are you sure it is installed in your environment? If so, look at the stack trace above to determine the origin of this error.",
                "See more at <https://github.com/openfisca/country-template#installing>.",
            ]
        )

        raise ImportError(message)
    if not hasattr(country_package, "CountryTaxBenefitSystem"):
        raise ImportError(
            "`{}` does not seem to be a valid Openfisca country package.".format(
                country_package_name
            )
        )

    country_package = importlib.import_module(country_package_name)
    tax_benefit_system = country_package.CountryTaxBenefitSystem()

    if extensions:
        for extension in extensions:
            tax_benefit_system.load_extension(extension)

    if reforms:
        for reform_path in reforms:
            tax_benefit_system = tax_benefit_system.apply_reform(reform_path)

    return tax_benefit_system


def detect_country_package():
    installed_country_packages = []
    for module_description in pkgutil.iter_modules():
        module_name = module_description[1]
        if "policyengine" in module_name.lower():
            try:
                module = importlib.import_module(module_name)
                if hasattr(module, "CountryTaxBenefitSystem"):
                    installed_country_packages.append(module_name)
            except ImportError:
                pass
    import policyengine_core.country_template as country_template

    installed_country_packages.append(country_template.__name__)

    if len(installed_country_packages) == 0:
        raise ImportError(
            "No country package has been detected on your environment. If your country package is installed but not detected, please use the --country-package option."
        )
    if len(installed_country_packages) > 1:
        log.warning(
            "Several country packages detected : `{}`. Using `{}` by default. To use another package, please use the --country-package option.".format(
                ", ".join(installed_country_packages),
                installed_country_packages[0],
            )
        )
    return installed_country_packages[0]



from typing import List

from policyengine_core.entities.entity import Entity
from policyengine_core.entities.role import Role


class GroupEntity(Entity):
    """Represents an entity containing several others with different roles.

    A :class:`.GroupEntity` represents an :class:`.Entity` containing
    several other :class:`.Entity` with different :class:`.Role`, and on
    which calculations can be run.

    Args:
        key: A key to identify the group entity.
        plural: The ``key``, pluralised.
        label: A summary description.
        doc: A full description.
        roles: The list of :class:`.Role` of the group entity.
        containing_entities: The list of keys of group entities whose members
            are guaranteed to be a superset of this group's entities.

    .. versionchanged:: 35.7.0
        Added ``containing_entities``, that allows the defining of group
        entities which entirely contain other group entities.

    """

    def __init__(
        self,
        key: str,
        plural: str,
        label: str,
        doc: str,
        roles: List[str],
        containing_entities: List[str] = (),
    ):
        super().__init__(key, plural, label, doc)
        self.roles_description = roles
        self.roles = []
        for role_description in roles:
            role = Role(role_description, self)
            setattr(self, role.key.upper(), role)
            self.roles.append(role)
            if role_description.get("subroles"):
                role.subroles = []
                for subrole_key in role_description["subroles"]:
                    subrole = Role({"key": subrole_key, "max": 1}, self)
                    setattr(self, subrole.key.upper(), subrole)
                    role.subroles.append(subrole)
                role.max = len(role.subroles)
        self.flattened_roles = sum(
            [role2.subroles or [role2] for role2 in self.roles], []
        )
        self.is_person = False
        self.containing_entities = containing_entities



from __future__ import annotations

import bisect
import itertools
import typing

import numpy

from policyengine_core import taxscales
from policyengine_core.taxscales.rate_tax_scale_like import RateTaxScaleLike

if typing.TYPE_CHECKING:
    NumericalArray = typing.Union[numpy.int_, numpy.float_]


class MarginalRateTaxScale(RateTaxScaleLike):
    def add_tax_scale(self, tax_scale: RateTaxScaleLike) -> None:
        # So as not to have problems with empty scales
        if len(tax_scale.thresholds) > 0:
            for threshold_low, threshold_high, rate in zip(
                tax_scale.thresholds[:-1],
                tax_scale.thresholds[1:],
                tax_scale.rates,
            ):
                self.combine_bracket(rate, threshold_low, threshold_high)

            # To process the last threshold
            self.combine_bracket(
                tax_scale.rates[-1],
                tax_scale.thresholds[-1],
            )

    def calc(
        self,
        tax_base: NumericalArray,
        factor: float = 1.0,
        round_base_decimals: typing.Optional[int] = None,
    ) -> numpy.float_:
        """
        Compute the tax amount for the given tax bases by applying a taxscale.

        :param ndarray tax_base: Array of the tax bases.
        :param float factor: Factor to apply to the thresholds of the taxscale.
        :param int round_base_decimals: Decimals to keep when rounding
                                        thresholds.

        :returns: Float array with tax amount for the given tax bases.

        For instance:

        >>> tax_scale = MarginalRateTaxScale()
        >>> tax_scale.add_bracket(0, 0)
        >>> tax_scale.add_bracket(100, 0.1)
        >>> tax_base = array([0, 150])
        >>> tax_scale.calc(tax_base)
        [0.0, 5.0]
        """
        base1 = numpy.tile(tax_base, (len(self.thresholds), 1)).T
        factor = numpy.ones(len(tax_base)) * factor

        # To avoid the creation of:
        #
        #   numpy.nan = 0 * numpy.inf
        #
        # We use:
        #
        #   numpy.finfo(float_).eps
        thresholds1 = numpy.outer(
            factor + numpy.finfo(numpy.float_).eps,
            numpy.array(self.thresholds + [numpy.inf]),
        )

        if round_base_decimals is not None:
            thresholds1 = numpy.round_(thresholds1, round_base_decimals)

        a = numpy.maximum(
            numpy.minimum(base1, thresholds1[:, 1:]) - thresholds1[:, :-1], 0
        )

        if round_base_decimals is None:
            return numpy.dot(self.rates, a.T)

        else:
            r = numpy.tile(self.rates, (len(tax_base), 1))
            b = numpy.round_(a, round_base_decimals)
            return numpy.round_(r * b, round_base_decimals).sum(axis=1)

    def combine_bracket(
        self,
        rate: typing.Union[int, float],
        threshold_low: int = 0,
        threshold_high: typing.Union[int, bool] = False,
    ) -> None:
        # Insert threshold_low and threshold_high without modifying rates
        if threshold_low not in self.thresholds:
            index = bisect.bisect_right(self.thresholds, threshold_low) - 1
            self.add_bracket(threshold_low, self.rates[index])

        if threshold_high and threshold_high not in self.thresholds:
            index = bisect.bisect_right(self.thresholds, threshold_high) - 1
            self.add_bracket(threshold_high, self.rates[index])

        # Use add_bracket to add rates where they belongs
        i = self.thresholds.index(threshold_low)

        if threshold_high:
            j = self.thresholds.index(threshold_high) - 1

        else:
            j = len(self.thresholds) - 1

        while i <= j:
            self.add_bracket(self.thresholds[i], rate)
            i += 1

    def marginal_rates(
        self,
        tax_base: NumericalArray,
        factor: float = 1.0,
        round_base_decimals: typing.Optional[int] = None,
    ) -> numpy.float_:
        """
        Compute the marginal tax rates relevant for the given tax bases.

        :param ndarray tax_base: Array of the tax bases.
        :param float factor: Factor to apply to the thresholds of a tax scale.
        :param int round_base_decimals: Decimals to keep when rounding
                                        thresholds.

        :returns: Float array with relevant marginal tax rate for the given tax
                  bases.

        For instance:

        >>> tax_scale = MarginalRateTaxScale()
        >>> tax_scale.add_bracket(0, 0)
        >>> tax_scale.add_bracket(100, 0.1)
        >>> tax_base = array([0, 150])
        >>> tax_scale.marginal_rates(tax_base)
        [0.0, 0.1]
        """
        bracket_indices = self.bracket_indices(
            tax_base,
            factor,
            round_base_decimals,
        )

        return numpy.array(self.rates)[bracket_indices]

    def rate_from_bracket_indice(
        self,
        bracket_indice: numpy.int_,
    ) -> numpy.float_:
        """
        Compute the relevant tax rates for the given bracket indices.

        :param: ndarray bracket_indice: Array of the bracket indices.

        :returns: Floating array with relevant tax rates
                  for the given bracket indices.

        For instance:

        >>> import numpy
        >>> tax_scale = MarginalRateTaxScale()
        >>> tax_scale.add_bracket(0, 0)
        >>> tax_scale.add_bracket(200, 0.1)
        >>> tax_scale.add_bracket(500, 0.25)
        >>> tax_base = numpy.array([50, 1_000, 250])
        >>> bracket_indice = tax_scale.bracket_indices(tax_base)
        >>> tax_scale.rate_from_bracket_indice(bracket_indice)
        array([0.  , 0.25, 0.1 ])
        """

        if bracket_indice.max() > len(self.rates) - 1:
            raise IndexError(
                f"bracket_indice parameter ({bracket_indice}) "
                f"contains one or more bracket indice which is unavailable "
                f"inside current {self.__class__.__name__} :\n"
                f"{self}"
            )

        return numpy.array(self.rates)[bracket_indice]

    def rate_from_tax_base(
        self,
        tax_base: NumericalArray,
    ) -> numpy.float_:
        """
        Compute the relevant tax rates for the given tax bases.

        :param: ndarray tax_base: Array of the tax bases.

        :returns: Floating array with relevant tax rates
                  for the given tax bases.

        For instance:

        >>> import numpy
        >>> tax_scale = MarginalRateTaxScale()
        >>> tax_scale.add_bracket(0, 0)
        >>> tax_scale.add_bracket(200, 0.1)
        >>> tax_scale.add_bracket(500, 0.25)
        >>> tax_base = numpy.array([1_000, 50, 450])
        >>> tax_scale.rate_from_tax_base(tax_base)
        array([0.25, 0.  , 0.1 ])
        """

        return self.rate_from_bracket_indice(self.bracket_indices(tax_base))

    def inverse(self) -> MarginalRateTaxScale:
        """
        Returns a new instance of MarginalRateTaxScale.

        Invert a taxscale:

            Assume tax_scale composed of bracket whose thresholds are expressed
            in terms of gross revenue.

            The inverse is another MarginalRateTaxScale whose thresholds are
            expressed in terms of net revenue.

            If net = gross_revenue - tax_scale.calc(gross_revenue)
            Then gross = tax_scale.inverse().calc(net)
        """
        # Threshold of net revenue.
        net_threshold: int = 0

        # Threshold of gross revenue.
        threshold: int

        # The intercept of the segments of the different thresholds in a
        # representation of taxable revenue as a piecewise linear function
        # of gross revenue.
        theta: int

        # Actually 1 / (1 - global_rate)
        inverse = self.__class__(
            name=str(self.name) + "'",
            option=self.option,
            unit=self.unit,
        )

        for threshold, rate in zip(self.thresholds, self.rates):
            if threshold == 0:
                previous_rate = 0
                theta = 0

            # We calculate the taxable revenue threshold of the considered
            # bracket.
            net_threshold = (1 - previous_rate) * threshold + theta
            inverse.add_bracket(net_threshold, 1 / (1 - rate))
            theta = (rate - previous_rate) * threshold + theta
            previous_rate = rate

        return inverse

    def scale_tax_scales(self, factor: float) -> MarginalRateTaxScale:
        """Scale all the MarginalRateTaxScales in the node."""
        scaled_tax_scale = self.copy()
        return scaled_tax_scale.multiply_thresholds(factor)

    def to_average(self) -> taxscales.LinearAverageRateTaxScale:
        average_tax_scale = taxscales.LinearAverageRateTaxScale(
            name=self.name,
            option=self.option,
            unit=self.unit,
        )

        average_tax_scale.add_bracket(0, 0)

        if self.thresholds:
            i = 0
            previous_threshold = self.thresholds[0]
            previous_rate = self.rates[0]

            for threshold, rate in itertools.islice(
                zip(self.thresholds, self.rates),
                1,
                None,
            ):
                i += previous_rate * (threshold - previous_threshold)
                average_tax_scale.add_bracket(threshold, i / threshold)
                previous_threshold = threshold
                previous_rate = rate

            average_tax_scale.add_bracket(float("Inf"), rate)

        return average_tax_scale



from typing import Optional

import numpy

from policyengine_core.types import ArrayLike, ArrayType


def average_rate(
    target: ArrayType[float],
    varying: ArrayLike[float],
    trim: Optional[ArrayLike[float]] = None,
) -> ArrayType[float]:
    """Computes the average rate of a target net income.

    Given a ``target`` net income, and according to the ``varying`` gross
    income. Optionally, a ``trim`` can be applied consisting of the lower and
    upper bounds of the average rate to be computed.

    Note:
        Usually, ``target`` and ``varying`` are the same size.

    Args:
        target: The targeted net income.
        varying: The varying gross income.
        trim: The lower and upper bounds of the average rate.

    Returns:
        :obj:`numpy.ndarray` of :obj:`float`:

        The average rate for each target.

        When ``trim`` is provided, values that are out of the provided bounds
        are replaced by :obj:`numpy.nan`.

    Examples:
        >>> target = numpy.array([1, 2, 3])
        >>> varying = [2, 2, 2]
        >>> trim = [-1, .25]
        >>> average_rate(target, varying, trim)
        array([ nan,  0. , -0.5])

    """

    average_rate: ArrayType[float]

    average_rate = 1 - target / varying

    if trim is not None:

        average_rate = numpy.where(
            average_rate <= max(trim),
            average_rate,
            numpy.nan,
        )

        average_rate = numpy.where(
            average_rate >= min(trim),
            average_rate,
            numpy.nan,
        )

    return average_rate


def marginal_rate(
    target: ArrayType[float],
    varying: ArrayType[float],
    trim: Optional[ArrayLike[float]] = None,
) -> ArrayType[float]:
    """Computes the marginal rate of a target net income.

    Given a ``target`` net income, and according to the ``varying`` gross
    income. Optionally, a ``trim`` can be applied consisting of the lower and
    upper bounds of the marginal rate to be computed.

    Note:
        Usually, ``target`` and ``varying`` are the same size.

    Args:
        target: The targeted net income.
        varying: The varying gross income.
        trim: The lower and upper bounds of the marginal rate.

    Returns:
        :obj:`numpy.ndarray` of :obj:`float`:

        The marginal rate for each target.

        When ``trim`` is provided, values that are out of the provided bounds
        are replaced by :obj:`numpy.nan`.

    Examples:
        >>> target = numpy.array([1, 2, 3])
        >>> varying = numpy.array([1, 2, 4])
        >>> trim = [.25, .75]
        >>> marginal_rate(target, varying, trim)
        array([nan, 0.5])

    """

    marginal_rate: ArrayType[float]

    marginal_rate = +1 - (target[:-1] - target[1:]) / (
        varying[:-1] - varying[1:]
    )

    if trim is not None:

        marginal_rate = numpy.where(
            marginal_rate <= max(trim),
            marginal_rate,
            numpy.nan,
        )

        marginal_rate = numpy.where(
            marginal_rate >= min(trim),
            marginal_rate,
            numpy.nan,
        )

    return marginal_rate



from typing import Any, Callable

import numpy as np
from numpy.typing import ArrayLike
from openfisca_core.entities import Entity
from openfisca_core.populations import GroupPopulation, Population
from openfisca_core.projectors import EntityToPersonProjector, Projector
from openfisca_core.variables import Variable


class CallableSubset:
    def __init__(
        self, population: Population, callable: Callable, mask: ArrayLike
    ):
        self.population = population
        self.callable = callable
        self.mask = mask

    def __call__(self, array: ArrayLike = None, *args, **kwargs):
        if array is None:
            return self.callable(*args, **kwargs)[self.mask]
        # Here, we're in e.g. household.sum(...).
        # The OpenFisca Population objects can map from e.g. people to households,
        # but they use primary/foreign key maps for the entire population. So,
        # for a subset of the population, we need to go back to the full population
        # (filling in with zeroes), use OpenFisca's mapping, then re-filter.
        if len(array) == self.population.members.count:
            return self.callable(array, *args, **kwargs)[self.mask]
        decompressed_size = self.population.members.count
        if len(self.mask) != decompressed_size:
            mask = self.population.project(self.mask)
        else:
            mask = self.mask
        decompressed_array = np.zeros((decompressed_size,))
        decompressed_array[mask] = array
        return self.callable(decompressed_array, *args, **kwargs)[self.mask]


class PopulationSubset:
    def __init__(self, population: Population, mask: ArrayLike):
        self.population = population
        self.mask = mask

    def __call__(self, *args, **kwargs):
        return self.population(*args, **kwargs)[self.mask]

    def __getattribute__(self, attribute):
        if attribute in ("population", "mask"):
            return object.__getattribute__(self, attribute)
        original_result = getattr(self.population, attribute)
        if isinstance(original_result, EntityToPersonProjector):
            # e.g. person.household
            return PopulationSubset(original_result, self.mask)
        elif attribute in (
            "sum",
            "min",
            "max",
            "nb_persons",
            "any",
            "all",
            "value_from_first_person",
        ):
            return CallableSubset(self.population, original_result, self.mask)
        return original_result


def make_partially_executed_formula(
    formula: Callable,
    mask: ArrayLike,
    default_value: Any = 0,
    value_type: type = float,
) -> Callable:
    # Edge cases that need to be covered:
    # * entity(variable, period)
    # * entity.members(variable, period)
    # * entity.parent_entity(variable, period)

    def partially_executed_formula(entity, period, parameters):
        if isinstance(mask, str):
            mask_entity = entity.simulation.tax_benefit_system.variables[
                mask
            ].entity.key
            if entity.entity.key != mask_entity:
                mask_values = getattr(entity, mask_entity)(mask, period)
            else:
                mask_values = entity(mask, period)
        else:
            mask_values = mask

        result = np.ones_like(mask_values, dtype=value_type) * default_value

        if not mask_values.any():
            return result

        subset_entity = PopulationSubset(entity, mask_values)

        formula_result = formula(subset_entity, period, parameters)
        formula_result = np.array(formula_result)

        result[mask_values] = formula_result

        entity = subset_entity.population

        return result

    return partially_executed_formula



from __future__ import annotations

import copy
from typing import Callable, Union

from policyengine_core.parameters import ParameterNode, Parameter
from policyengine_core.taxbenefitsystems import TaxBenefitSystem


class Reform(TaxBenefitSystem):
    """A modified TaxBenefitSystem

    All reforms must subclass `Reform` and implement a method `apply()`.

    In this method, the reform can add or replace variables and call `modify_parameters` to modify the parameters of the legislation.

        Example:

        >>> from policyengine_core import reforms
        >>> from policyengine_core.parameters import load_parameter_file
        >>>
        >>> def modify_my_parameters(parameters):
        >>>     # Add new parameters
        >>>     new_parameters = load_parameter_file(name='reform_name', file_path='path_to_yaml_file.yaml')
        >>>     parameters.add_child('reform_name', new_parameters)
        >>>
        >>>     # Update a value
        >>>     parameters.taxes.some_tax.some_param.update(period=some_period, value=1000.0)
        >>>
        >>>    return parameters
        >>>
        >>> class MyReform(reforms.Reform):
        >>>    def apply(self):
        >>>        self.add_variable(some_variable)
        >>>        self.update_variable(some_other_variable)
        >>>        self.modify_parameters(modifier_function = modify_my_parameters)
    """

    name: str = None

    def __init__(self, baseline: TaxBenefitSystem):
        """
        :param baseline: Baseline TaxBenefitSystem.
        """
        super().__init__(baseline.entities)
        self.baseline = baseline
        self.parameters = baseline.parameters
        self._parameters_at_instant_cache = (
            baseline._parameters_at_instant_cache
        )
        self.variables = baseline.variables.copy()
        self.decomposition_file_path = baseline.decomposition_file_path
        self.key = self.__class__.__name__
        if not hasattr(self, "apply"):
            raise Exception(
                "Reform {} must define an `apply` function".format(self.key)
            )
        self.apply()

    def __getattr__(self, attribute):
        return getattr(self.baseline, attribute)

    @property
    def full_key(self) -> str:
        key = self.key
        assert (
            key is not None
        ), "key was not set for reform {} (name: {!r})".format(self, self.name)
        if self.baseline is not None and hasattr(self.baseline, "key"):
            baseline_full_key = self.baseline.full_key
            key = ".".join([baseline_full_key, key])
        return key

    def modify_parameters(
        self, modifier_function: Callable[[ParameterNode], ParameterNode]
    ) -> None:
        """Make modifications on the parameters of the legislation.

        Call this function in `apply()` if the reform asks for legislation parameter modifications.

        Args:
            modifier_function: A function that takes a :obj:`.ParameterNode` and should return an object of the same type.
        """
        baseline_parameters = self.baseline.parameters
        baseline_parameters_copy = copy.deepcopy(baseline_parameters)
        reform_parameters = modifier_function(baseline_parameters_copy)
        if not isinstance(reform_parameters, ParameterNode):
            return ValueError(
                "modifier_function {} in module {} must return a ParameterNode".format(
                    modifier_function.__name__,
                    modifier_function.__module__,
                )
            )
        self.parameters = reform_parameters
        self._parameters_at_instant_cache = {}


def set_parameter(
    path: Union[Parameter, str],
    value: float,
    period: str = "year:2015:10",
    return_modifier=False,
) -> Reform:
    if isinstance(path, Parameter):
        path = path.name

    def modifier(parameters: ParameterNode):
        node = parameters
        for name in path.split("."):
            try:
                if "[" not in name:
                    node = node.children[name]
                else:
                    try:
                        name, index = name.split("[")
                        index = int(index[:-1])
                        node = node.children[name].brackets[index]
                    except:
                        raise ValueError(
                            "Invalid bracket syntax (should be e.g. tax.brackets[3].rate"
                        )
            except:
                raise ValueError(
                    f"Could not find the parameter (failed at {name})."
                )
        node.update(period=period, value=value)
        return parameters

    if return_modifier:
        return modifier

    class reform(Reform):
        def apply(self):
            self.modify_parameters(modifier)

    return reform



# -*- coding: utf-8 -*-


import os

import numpy as np

from policyengine_core.data_storage import OnDiskStorage
from policyengine_core.periods import ETERNITY
from policyengine_core.simulations import Simulation


def dump_simulation(simulation, directory):
    """
    Write simulation data to directory, so that it can be restored later.
    """
    parent_directory = os.path.abspath(os.path.join(directory, os.pardir))
    if not os.path.isdir(parent_directory):  # To deal with reforms
        os.mkdir(parent_directory)
    if not os.path.isdir(directory):
        os.mkdir(directory)

    if os.listdir(directory):
        raise ValueError("Directory '{}' is not empty".format(directory))

    entities_dump_dir = os.path.join(directory, "__entities__")
    os.mkdir(entities_dump_dir)

    for entity in simulation.populations.values():
        # Dump entity structure
        _dump_entity(entity, entities_dump_dir)

        # Dump variable values
        for holder in entity._holders.values():
            _dump_holder(holder, directory)


def restore_simulation(directory, tax_benefit_system, **kwargs):
    """
    Restore simulation from directory
    """
    simulation = Simulation(
        tax_benefit_system, tax_benefit_system.instantiate_entities()
    )

    entities_dump_dir = os.path.join(directory, "__entities__")
    for population in simulation.populations.values():
        if population.entity.is_person:
            continue
        person_count = _restore_entity(population, entities_dump_dir)

    for population in simulation.populations.values():
        if not population.entity.is_person:
            continue
        _restore_entity(population, entities_dump_dir)
        population.count = person_count

    variables_to_restore = (
        variable
        for variable in os.listdir(directory)
        if variable != "__entities__"
    )
    for variable in variables_to_restore:
        _restore_holder(simulation, variable, directory)

    return simulation


def _dump_holder(holder, directory):
    disk_storage = holder.create_disk_storage(directory, preserve=True)
    for period in holder.get_known_periods():
        value = holder.get_array(period)
        disk_storage.put(value, period)


def _dump_entity(population, directory):
    path = os.path.join(directory, population.entity.key)
    os.mkdir(path)
    np.save(os.path.join(path, "id.npy"), population.ids)

    if population.entity.is_person:
        return

    np.save(
        os.path.join(path, "members_position.npy"), population.members_position
    )
    np.save(
        os.path.join(path, "members_entity_id.npy"),
        population.members_entity_id,
    )

    flattened_roles = population.entity.flattened_roles
    if len(flattened_roles) == 0:
        encoded_roles = np.int64(0)
    else:
        encoded_roles = np.select(
            [population.members_role == role for role in flattened_roles],
            [role.key for role in flattened_roles],
        )
    np.save(os.path.join(path, "members_role.npy"), encoded_roles)


def _restore_entity(population, directory):
    path = os.path.join(directory, population.entity.key)

    population.ids = np.load(os.path.join(path, "id.npy"))

    if population.entity.is_person:
        return

    population.members_position = np.load(
        os.path.join(path, "members_position.npy")
    )
    population.members_entity_id = np.load(
        os.path.join(path, "members_entity_id.npy")
    )
    encoded_roles = np.load(os.path.join(path, "members_role.npy"))

    flattened_roles = population.entity.flattened_roles
    if len(flattened_roles) == 0:
        population.members_role = np.int64(0)
    else:
        population.members_role = np.select(
            [encoded_roles == role.key for role in flattened_roles],
            [role for role in flattened_roles],
        )
    person_count = len(population.members_entity_id)
    population.count = max(population.members_entity_id) + 1
    return person_count


def _restore_holder(simulation, variable, directory):
    storage_dir = os.path.join(directory, variable)
    is_variable_eternal = (
        simulation.tax_benefit_system.get_variable(variable).definition_period
        == ETERNITY
    )
    disk_storage = OnDiskStorage(
        storage_dir, is_eternal=is_variable_eternal, preserve_storage_dir=True
    )
    disk_storage.restore()

    holder = simulation.get_holder(variable)

    for period in disk_storage.get_known_periods():
        value = disk_storage.get(period)
        holder.put_in_cache(value, period)



import logging

import numpy
from numpy.typing import ArrayLike

from policyengine_core import periods
from policyengine_core.holders.holder import Holder
from policyengine_core.periods import Period

log = logging.getLogger(__name__)


def set_input_dispatch_by_period(
    holder: Holder, period: Period, array: ArrayLike
):
    """
    This function can be declared as a ``set_input`` attribute of a variable.

    In this case, the variable will accept inputs on larger periods that its definition period, and the value for the larger period will be applied to all its subperiods.

    To read more about ``set_input`` attributes, check the `documentation <https://openfisca.org/doc/coding-the-legislation/35_periods.html#set-input-automatically-process-variable-inputs-defined-for-periods-not-matching-the-definition-period>`_.
    """
    array = holder._to_array(array)

    period_size = period.size
    period_unit = period.unit

    if holder.variable.definition_period == periods.MONTH:
        cached_period_unit = periods.MONTH
    elif holder.variable.definition_period == periods.YEAR:
        cached_period_unit = periods.YEAR
    else:
        raise ValueError(
            "set_input_dispatch_by_period can be used only for yearly or monthly variables."
        )

    after_instant = period.start.offset(period_size, period_unit)

    # Cache the input data, skipping the existing cached months
    sub_period = period.start.period(cached_period_unit)
    while sub_period.start < after_instant:
        existing_array = holder.get_array(sub_period)
        if existing_array is None:
            holder._set(sub_period, array)
        else:
            # The array of the current sub-period is reused for the next ones.
            # TODO: refactor or document this behavior
            array = existing_array
        sub_period = sub_period.offset(1)


def set_input_divide_by_period(
    holder: Holder, period: Period, array: ArrayLike
):
    """
    This function can be declared as a ``set_input`` attribute of a variable.

    In this case, the variable will accept inputs on larger periods that its definition period, and the value for the larger period will be divided between its subperiods.

    To read more about ``set_input`` attributes, check the `documentation <https://openfisca.org/doc/coding-the-legislation/35_periods.html#set-input-automatically-process-variable-inputs-defined-for-periods-not-matching-the-definition-period>`_.
    """
    if not isinstance(array, numpy.ndarray):
        array = numpy.array(array)
    period_size = period.size
    period_unit = period.unit

    if holder.variable.definition_period == periods.MONTH:
        cached_period_unit = periods.MONTH
    elif holder.variable.definition_period == periods.YEAR:
        cached_period_unit = periods.YEAR
    else:
        raise ValueError(
            "set_input_divide_by_period can be used only for yearly or monthly variables."
        )

    after_instant = period.start.offset(period_size, period_unit)

    # Count the number of elementary periods to change, and the difference with what is already known.
    remaining_array = array.copy()
    sub_period = period.start.period(cached_period_unit)
    sub_periods_count = 0
    while sub_period.start < after_instant:
        existing_array = holder.get_array(sub_period)
        if existing_array is not None:
            remaining_array -= existing_array
        else:
            sub_periods_count += 1
        sub_period = sub_period.offset(1)

    # Cache the input data
    if sub_periods_count > 0:
        divided_array = remaining_array / sub_periods_count
        sub_period = period.start.period(cached_period_unit)
        while sub_period.start < after_instant:
            if holder.get_array(sub_period) is None:
                holder._set(sub_period, divided_array)
            sub_period = sub_period.offset(1)
    elif not (remaining_array == 0).all():
        raise ValueError(
            "Inconsistent input: variable {0} has already been set for all months contained in period {1}, and value {2} provided for {1} doesn't match the total ({3}). This error may also be thrown if you try to call set_input twice for the same variable and period.".format(
                holder.variable.name, period, array, array - remaining_array
            )
        )



from typing import Sequence, TypeVar, Union

import numpy
from nptyping import NDArray as ArrayType
from nptyping import types

T = TypeVar("T", bool, bytes, float, int, object, str)

types._ndarray_meta._Type = Union[type, numpy.dtype, TypeVar]

ArrayLike = Union[ArrayType[T], Sequence[T]]
""":obj:`typing.Generic`: Type of any castable to :class:`numpy.ndarray`.

These include any :obj:`numpy.ndarray` and sequences (like
:obj:`list`, :obj:`tuple`, and so on).

Examples:
    >>> ArrayLike[float]
    typing.Union[numpy.ndarray, typing.Sequence[float]]

    >>> ArrayLike[str]
    typing.Union[numpy.ndarray, typing.Sequence[str]]

Note:
    It is possible since numpy version 1.21 to specify the type of an
    array, thanks to `numpy.typing.NDArray`_::

        from numpy.typing import NDArray
        NDArray[numpy.float64]

    `mypy`_ provides `duck type compatibility`_, so an :obj:`int` is
    considered to be valid whenever a :obj:`float` is expected.

Todo:
    * Refactor once numpy version >= 1.21 is used.

.. versionadded:: 35.5.0

.. versionchanged:: 35.6.0
    Moved to :mod:`.types`

.. _mypy:
    https://mypy.readthedocs.io/en/stable/

.. _duck type compatibility:
    https://mypy.readthedocs.io/en/stable/duck_type_compatibility.html

.. _numpy.typing.NDArray:
    https://numpy.org/doc/stable/reference/typing.html#numpy.typing.NDArray

"""



"""
This file defines a reform.

A reform is a set of modifications to be applied to a reference tax and benefit system to carry out experiments.

See https://openfisca.org/doc/key-concepts/reforms.html
"""

# Import from openfisca-core the Python objects used to code the legislation in OpenFisca
from policyengine_core.reforms import Reform
from policyengine_core.variables import Variable


class social_security_contribution(Variable):
    # Variable metadata don't need to be redefined. By default, the reference variable metadatas will be used.

    def formula(person, period, _parameters):
        """
        Social security contribution reform.

        Our reform replaces `social_security_contribution` (the "reference" variable) by the following variable.
        """
        return person("salary", period) * 0.03


class flat_social_security_contribution(Reform):
    def apply(self):
        """
        Apply reform.

        A reform always defines an `apply` method that builds the reformed tax and benefit system from the reference one.
        See https://openfisca.org/doc/coding-the-legislation/reforms.html#writing-a-reform
        """
        self.update_variable(social_security_contribution)



from numpy import ceil, floor

from policyengine_core.parameters.operations.get_parameter import get_parameter
from policyengine_core.parameters.parameter import Parameter
from policyengine_core.parameters.parameter_at_instant import (
    ParameterAtInstant,
)
from policyengine_core.parameters.parameter_node import ParameterNode
from policyengine_core.parameters.parameter_scale import ParameterScale
from policyengine_core.periods import instant, period


def uprate_parameters(root: ParameterNode) -> ParameterNode:
    """Uprates parameters according to their metadata.

    Args:
        root (ParameterNode): The root of the parameter tree.

    Returns:
        ParameterNode: The same root, with uprating applied to descendants.
    """

    descendants = list(root.get_descendants())

    scales = list(filter(lambda p: isinstance(p, ParameterScale), descendants))
    for scale in scales:
        for bracket in scale.brackets:
            for allowed_key in bracket._allowed_keys:
                if hasattr(bracket, allowed_key):
                    descendants.append(getattr(bracket, allowed_key))

    for parameter in descendants:
        if isinstance(parameter, Parameter):
            if parameter.metadata.get("uprating") is not None:
                meta = parameter.metadata["uprating"]
                if meta == "self":
                    meta = dict(parameter="self")
                elif isinstance(meta, str):
                    meta = dict(parameter=meta)
                if meta["parameter"] == "self":
                    last_instant = instant(
                        parameter.values_list[0].instant_str
                    )
                    start_instant = instant(
                        meta.get(
                            "from",
                            last_instant.offset(
                                -1, meta.get("interval", "year")
                            ),
                        )
                    )
                    start = parameter(start_instant)
                    end_instant = instant(meta.get("to", last_instant))
                    end = parameter(end_instant)
                    increase = end / start
                    if "from" in meta:
                        # This won't work for non-year periods, which are more complicated
                        increase / (end_instant.year - start_instant.year)
                    data = {}
                    value = parameter.values_list[0].value
                    for i in range(meta.get("number", 10)):
                        data[
                            str(
                                last_instant.offset(
                                    i, meta.get("interval", "year")
                                )
                            )
                        ] = (value * increase)
                        value *= increase
                    uprating_parameter = Parameter("self", data=data)
                else:
                    uprating_parameter = get_parameter(root, meta["parameter"])
                # Start from the latest value
                if "start_instant" in meta:
                    last_instant = instant(meta["start_instant"])
                else:
                    last_instant = instant(
                        parameter.values_list[0].instant_str
                    )
                # For each defined instant in the uprating parameter
                for entry in uprating_parameter.values_list[::-1]:
                    entry_instant = instant(entry.instant_str)
                    # If the uprater instant is defined after the last parameter instant
                    if entry_instant > last_instant:
                        # Apply the uprater and add to the parameter
                        value_at_start = parameter(last_instant)
                        uprater_at_start = uprating_parameter(last_instant)
                        uprater_at_entry = uprating_parameter(entry_instant)
                        uprater_change = uprater_at_entry / uprater_at_start
                        uprated_value = value_at_start * uprater_change
                        if "rounding" in meta:
                            rounding_config = meta["rounding"]
                            if isinstance(rounding_config, float):
                                interval = rounding_config
                                rounding_fn = round
                            elif isinstance(rounding_config, dict):
                                interval = rounding_config["interval"]
                                rounding_fn = dict(
                                    nearest=round,
                                    upwards=ceil,
                                    downwards=floor,
                                )[rounding_config["type"]]
                            uprated_value = (
                                rounding_fn(uprated_value / interval)
                                * interval
                            )
                        parameter.values_list.append(
                            ParameterAtInstant(
                                parameter.name,
                                entry.instant_str,
                                data=uprated_value,
                            )
                        )
                parameter.values_list.sort(
                    key=lambda x: x.instant_str, reverse=True
                )
    return root



# -*- coding: utf-8 -*-


import os

import numexpr

from policyengine_core.enums import EnumArray


def assert_near(
    value,
    target_value,
    absolute_error_margin=None,
    message="",
    relative_error_margin=None,
):
    """

    :param value: Value returned by the test
    :param target_value: Value that the test should return to pass
    :param absolute_error_margin: Absolute error margin authorized
    :param message: Error message to be displayed if the test fails
    :param relative_error_margin: Relative error margin authorized

    Limit : This function cannot be used to assert near periods.

    """

    import numpy as np

    if absolute_error_margin is None and relative_error_margin is None:
        absolute_error_margin = 0
    if not isinstance(value, np.ndarray):
        value = np.array(value)
    if isinstance(value, EnumArray):
        return assert_enum_equals(value, target_value, message)
    if np.issubdtype(value.dtype, np.datetime64):
        target_value = np.array(target_value, dtype=value.dtype)
        assert_datetime_equals(value, target_value, message)
    if isinstance(target_value, str):
        target_value = eval_expression(target_value)

    target_value = np.array(target_value).astype(np.float32)

    value = np.array(value).astype(np.float32)
    diff = abs(target_value - value)
    if absolute_error_margin is not None:
        assert (
            diff <= absolute_error_margin
        ).all(), "{}{} differs from {} with an absolute margin {} > {}".format(
            message, value, target_value, diff, absolute_error_margin
        )
    if relative_error_margin is not None:
        assert (
            diff <= abs(relative_error_margin * target_value)
        ).all(), "{}{} differs from {} with a relative margin {} > {}".format(
            message,
            value,
            target_value,
            diff,
            abs(relative_error_margin * target_value),
        )


def assert_datetime_equals(value, target_value, message=""):
    assert (value == target_value).all(), "{}{} differs from {}.".format(
        message, value, target_value
    )


def assert_enum_equals(value, target_value, message=""):
    value = value.decode_to_str()
    assert (value == target_value).all(), "{}{} differs from {}.".format(
        message, value, target_value
    )


def indent(text):
    return "  {}".format(text.replace(os.linesep, "{}  ".format(os.linesep)))


def get_trace_tool_link(scenario, variables, api_url, trace_tool_url):
    import json
    import urllib

    scenario_json = scenario.to_json()
    simulation_json = {
        "scenarios": [scenario_json],
        "variables": variables,
    }
    url = (
        trace_tool_url
        + "?"
        + urllib.urlencode(
            {
                "simulation": json.dumps(simulation_json),
                "api_url": api_url,
            }
        )
    )
    return url


def eval_expression(expression):
    try:
        return numexpr.evaluate(expression)
    except (KeyError, TypeError):
        return expression



"""
This file defines variables for the modelled legislation.

A variable is a property of an Entity such as a Person, a Household…

See https://openfisca.org/doc/key-concepts/variables.html
"""

# Import the Entities specifically defined for this tax and benefit system
from policyengine_core.country_template.entities import Household

# Import from openfisca-core the Python objects used to code the legislation in OpenFisca
from policyengine_core.enums import Enum
from policyengine_core.periods import MONTH
from policyengine_core.variables import Variable


# This variable is a pure input: it doesn't have a formula
class accommodation_size(Variable):
    value_type = float
    entity = Household
    definition_period = MONTH
    label = "Size of the accommodation, in square metres"


# This variable is a pure input: it doesn't have a formula
class rent(Variable):
    value_type = float
    entity = Household
    definition_period = MONTH
    label = "Rent paid by the household"


# Possible values for the housing_occupancy_status variable, defined further down
# See more at <https://openfisca.org/doc/coding-the-legislation/20_input_variables.html#advanced-example-enumerations-enum>
class HousingOccupancyStatus(Enum):
    __order__ = "owner tenant free_lodger homeless"
    owner = "Owner"
    tenant = "Tenant"
    free_lodger = "Free lodger"
    homeless = "Homeless"


class housing_occupancy_status(Variable):
    value_type = Enum
    possible_values = HousingOccupancyStatus
    default_value = HousingOccupancyStatus.tenant
    entity = Household
    definition_period = MONTH
    label = "Legal housing situation of the household concerning their main residence"


class postal_code(Variable):
    value_type = str
    max_length = 5
    entity = Household
    definition_period = MONTH
    label = "Postal code of the household"
